package httpclient

import (
	"bytes"
	"context"
	"crypto/tls"
	"fmt"
	"io/ioutil"
	"net"
	"net/http"
	"net/url"
	"strings"
	"time"

	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpproto"
	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// NSTAHTTPClient handles http requests to the master
type NSTAHTTPClient struct {
	ctx         context.Context
	endpoint    string
	satelliteID int
	apiKey      string
	client      http.Client
}

// New creates an nsta http client
func New(ctx context.Context, endpoint string, satelliteID int, apiKey string, proxy string, verify bool) (*NSTAHTTPClient, error) {
	var proxyFunc func(*http.Request) (*url.URL, error) = nil
	if proxy != "" {
		proxyURL, err := url.Parse(proxy)
		if err != nil {
			return nil, err
		}
		proxyFunc = http.ProxyURL(proxyURL)
	}

	endpoint = strings.TrimRight(endpoint, "/")

	// some settings from http.Transport were copied from the http.RoundTripper
	return &NSTAHTTPClient{
		ctx:         ctx,
		endpoint:    endpoint,
		satelliteID: satelliteID,
		apiKey:      apiKey,
		client: http.Client{
			Timeout: 5 * time.Minute,
			Transport: &http.Transport{
				ForceAttemptHTTP2:     true,
				MaxIdleConns:          100,
				IdleConnTimeout:       90 * time.Second,
				TLSHandshakeTimeout:   10 * time.Second,
				ExpectContinueTimeout: 1 * time.Second,
				DisableKeepAlives:     true,
				TLSClientConfig: &tls.Config{
					InsecureSkipVerify: !verify,
				},
				Proxy: proxyFunc,
				DialContext: (&net.Dialer{
					Timeout:   30 * time.Second,
					KeepAlive: 30 * time.Second,
				}).DialContext,
			},
		},
	}, nil
}

// SendMessages tries to send the gearman messages via the endpoint to the master
func (c *NSTAHTTPClient) SendMessages(msgs []*gearman.Message, lastExport time.Time) (*httpproto.HTTPGearmanMessage, error) {
	data, err := json.Marshal(httpproto.HTTPGearmanMessage{SatelliteID: c.satelliteID, APIKey: c.apiKey, Messages: msgs, LastExport: lastExport})
	if err != nil {
		return nil, err
	}

	req, err := http.NewRequestWithContext(c.ctx, "POST", fmt.Sprintf("%s/nsta/api/satcom", c.endpoint), bytes.NewBuffer(data))
	if err != nil {
		return nil, err
	}
	req.Header.Add("Content-Type", "application/json")
	resp, err := c.client.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("could not send http message, status %d: %s", resp.StatusCode, resp.Status)
	}
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return nil, fmt.Errorf("could not read body from server: %s", err)
	}

	httpResp := &httpproto.HTTPGearmanMessage{}
	err = json.Unmarshal(body, httpResp)

	return httpResp, err
}

// ExportExchange tries to send the latest export archive to the satellite
// keep tardata empty to request an export
func (c *NSTAHTTPClient) ExportExchange(lastExport time.Time, tardata string) (*httpproto.ExportData, error) {
	data, err := json.Marshal(httpproto.ExportData{SatelliteID: c.satelliteID, APIKey: c.apiKey, LastExport: lastExport, TARBase64: tardata})
	if err != nil {
		return nil, err
	}

	req, err := http.NewRequestWithContext(c.ctx, "POST", fmt.Sprintf("%s/nsta/api/export", c.endpoint), bytes.NewBuffer(data))
	if err != nil {
		return nil, err
	}
	req.Header.Add("Content-Type", "application/json")
	resp, err := c.client.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()
	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("could not send http message, status %d: %s", resp.StatusCode, resp.Status)
	}
	body, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		return nil, fmt.Errorf("could not read body from server: %s", err)
	}

	httpResp := &httpproto.ExportData{}
	err = json.Unmarshal(body, httpResp)

	return httpResp, err
}
